import socket

TCP_IP = '192.168.0.1'
# has probably always the same address
TCP_PORT = 9999
RECV_BYTE_SIZE = 2048


class TcpClient(object):
    """
    opens a socket for every object and close it after using
    an object of the class TcpClient can call the methods send or recv for TCP-communication
    """

    def __init__(self):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._socket.connect((TCP_IP, TCP_PORT))

    def __del__(self):
        self._socket.close()

    def send(self, encoded_message):
        self._socket.send(encoded_message)

    def recv(self):
        recv_bytes = self._socket.recv(RECV_BYTE_SIZE)
        return recv_bytes



from wifimeter.util.model import JSONResponse, EMeterData
import sys


def output_interface(raw_bytes):
    """
    Prints the measurement on screen
    :param raw_bytes: byteArray from the WlanMeter
    """
    response = JSONResponse.from_encoded(raw_bytes)
    measurement = EMeterData.from_json(response.json)
    print("{}, {}, {}, {}, {}, {}".format(measurement.date, measurement.current, measurement.total,
                                          measurement.power, measurement.voltage, measurement.err_code))


class Unbuffered(object):
    """
    Unbuffered stdout to allow working with the output easily (grep, more,...)
    http://stackoverflow.com/questions/107705/disable-output-buffering
    """

    def __init__(self, stream):
        self.stream = stream

    def write(self, data):
        # TODO: "python -m wifimeter measure | grep TestPlug" throws BrokenPipeError in ctrl+c
        self.stream.write(data)
        self.stream.flush()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)


sys.stdout = Unbuffered(sys.stdout)



import json
from datetime import datetime
from time import tzname
import pytz
from wifimeter.util.coder import decode, encode


class JsonSerializer(object):
    """
    Converts all self.variable of its childes to Json-keys and their values to values in Json
    """
    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, separators=(',', ':'), sort_keys=True)


class SetTimeRequest(JsonSerializer):
    """
    Sets the current time in format json. This entry can be sent to the device for its setup
    """
    def __init__(self):
        date = datetime.now(tz=pytz.timezone(tzname[0]))
        self.time = {
            'set_timezone': {
                'year': date.year,
                'month': date.month,
                'mday': date.day,
                'hour': date.hour,
                'min': date.minute,
                'sec': date.second,
                'index': 0
            }
        }

    @staticmethod
    def validate_response(response):
        """
        Checks the response of the device
        :param response: json-string
        """
        try:
            return response['time']['set_timezone']['err_code'] == 0
        except KeyError:
            return False


class SetAliasRequest(JsonSerializer):
    """
    Sets the alias in format json. This entry can be sent to the device for its setup
    """
    def __init__(self, alias):
        self.system = {
            'set_dev_alias': {
                'alias': alias
            }
        }

    @staticmethod
    def validate_response(response):
        """
        Checks the response of the device
        :param response: json-string
        """
        try:
            return response['system']['set_dev_alias']['err_code'] == 0
        except KeyError:
            return False


class SetScheduleRequest(JsonSerializer):
    """
    Resets the schedule (format json). This entry can be sent to the device for its setup
    """
    def __init__(self):
        self.schedule = {
            'get_next_action': None
        }

    @staticmethod
    def validate_response(response):
        """
        Checks the response of the device
        :param response: json-string
        """
        try:
            return response['schedule']['get_next_action']['err_code'] == 0
        except KeyError:
            return False


class SetNetifRequest(JsonSerializer):
    """
    Sets the wlan-name, password and type in format json. This entry can be sent to the device for its setup
    """
    def __init__(self, ssid, password, key_type):
        self.netif = {
            'set_stainfo': {
                'ssid': ssid,
                'password': password,
                'key_type': key_type
            }
        }

    @staticmethod
    def validate_response(response):
        """
        Checks the response of the device
        :param response: json-string
        """
        try:
            return response['netif']['set_stainfo']['err_code'] == 0
        except KeyError:
            return False


class NetifRefresher(JsonSerializer):
    """
    Write a refresh for the wlan in format json. This entry can be sent to the device for its setup
    """
    def __init__(self):
        self.netif = {
            'get_scaninfo': {
                'refresh': 0
            }
        }

    @staticmethod
    def validate_response(response):
        """
        Checks the response of the device
        :param response: json-string
        """
        try:
            return response['netif']['get_scaninfo']['err_code'] == 0
        except KeyError:
            return False


class MeasurementRequest(JsonSerializer):
    """
    Writes a json entry which can be sent to the device to get measurement
    """
    def __init__(self, state):
        self.system = {
                'get_sysinfo': state
        }
        self.emeter = JsonSerializer()
        self.emeter.get_realtime = None
        super(MeasurementRequest, self).__init__()


class RelayStateRequest(JsonSerializer):
    """
    Writes a json entry which can be sent to the device to get the current state of the device
    """
    def __init__(self, state):
        self.system = {
                'set_relay_state': {
                    'state': state
                }
        }
        super(RelayStateRequest, self).__init__()


class EMeterData(object):
    """
    Gets a Json-String and manages the single measurement.
    With the @property-functions is it possible to get the values
    """

    def __init__(self, date, current, total, power, voltage, err_code, alias):
        self._date = date
        self._current = current
        self._total = total
        self._power = power
        self._voltage = voltage
        self._err_code = err_code
        self._alias = alias

    @staticmethod
    def from_json(json_str):
        date = datetime.now(tz=pytz.timezone(tzname[0]))
        json_str = json.loads(json_str)
        measurement = json_str['emeter']['get_realtime']
        alias = json_str['system']['get_sysinfo']['alias']
        return EMeterData(alias,
                          date,
                          measurement['current'],
                          measurement['total'],
                          measurement['power'],
                          measurement['voltage'],
                          measurement['err_code'])

    @property
    def date(self):
        return self._date

    @property
    def current(self):
        return self._current

    @property
    def total(self):
        return self._total

    @property
    def power(self):
        return self._power

    @property
    def voltage(self):
        return self._voltage

    @property
    def err_code(self):
        return self._err_code

    @property
    def alias(self):
        return self._alias


class JSONResponse(object):
    def __init__(self, decoded_str):
        self._decoded_str = decoded_str

    @staticmethod
    def from_encoded(recv_bytes):
        return JSONResponse(decode(recv_bytes))

    @property
    def json(self):
        return self._decoded_str


class JSONRequest(object):
    def __init__(self, encoded_bytes):
        self._encoded_bytes = encoded_bytes

    @staticmethod
    def from_decoded(message_str):
        return JSONRequest(encode(message_str))

    @property
    def byte_array(self):
        return self._encoded_bytes



CODING_START_CHAR = 171

try:
    from itertools import izip
except ImportError:
    izip = zip


def decode(recv_bytes):
    """
    decodes the received byte_array to a string
    :param recv_bytes: byte_array; received from the device
    """
    recv_bytes = bytearray(recv_bytes)
    first_symbol = chr(CODING_START_CHAR)
    first_symbol = chr(recv_bytes[0] ^ ord(first_symbol))
    decoded_array = [chr(x ^ y) for x, y in izip(recv_bytes, recv_bytes[1:])]
    return first_symbol + "".join(decoded_array)


def encode(message_str):
    """
    encodes the message to a byte_array
    :param message_str: string in format json
    """
    ord_array = map(ord, message_str)
    message_byte = [CODING_START_CHAR]
    for c, s in izip(message_byte, ord_array):
        message_byte.append(c ^ s)
    return bytearray(message_byte[1:])



from getpass import getpass

import click
import logging
from wifimeter.network.threads import UDPSendThread, UDPRecvThread
from wifimeter.network.udp import UdpClient
from wifimeter.util.coder import encode
from wifimeter.util.model import MeasurementRequest, RelayStateRequest
from wifimeter.util.setup import tcp_setup
import timeout_decorator
import signal


@click.group()
def cli():
    pass


@cli.command()
@click.option('--state', prompt='State (0=OFF / 1=ON )', type=click.Choice(['0', '1']))
def switch(state):
    """
    Switch device ON/OFF
    """
    message_str = RelayStateRequest(int(state)).to_json()
    udp = UdpClient()
    udp.send(encode(message_str))


@cli.command()
def measure():
    """
    Receives readings every second
    """
    print("alias, timestamp, current, total, power, voltage, err_code")
    message_str = MeasurementRequest(None).to_json()
    socket_object = UdpClient()
    s = UDPSendThread(message_str, socket_object)
    r = UDPRecvThread(socket_object)
    s.start()
    r.start()

    def handler(signal, frame):
        for t in [s, r]:
            t.stop()
            t.join(1)
        socket_object.close()
        exit(0)

    signal.signal(signal.SIGINT, handler)

    while True:
        signal.pause()


@cli.command()
@click.option('-a', '--alias')
@click.option('-n', '--wlan-name')
@click.option('-p', '--password')
@click.option('-t', '--wlan-type', type=click.Choice(['0', '1', '2', '3']), help='0 = without any security, 2 = WEP, 3 = WPA')
@click.option('-l', '--log-level', default='DEBUG')
@timeout_decorator.timeout(35)
def setup(wlan_type, wlan_name, alias, password, log_level):
    """
    Setup device (alias, WLAN)
    """
    if alias is None:
        alias = click.prompt('Alias')
    if wlan_name is None:
        wlan_name = click.prompt("Wlan_name")
    if wlan_type is None:
        wlan_type = click.prompt("Wlan-type", type=click.Choice(['0', '1', '2', '3']))
    if wlan_type != '0' and password is None:
        password = getpass()
    setup_logging(log_level)
    wlan_type = int(wlan_type)
    tcp_setup(wlan_type, wlan_name, alias, password)


def setup_logging(log_level):
    """
    Set logging level
    :param log_level: possible choices 'DEBUG', 'WARNING', 'INFO'...
    """
    ext_logger = logging.getLogger("py.warnings")
    logging.captureWarnings(True)
    level = getattr(logging, log_level)
    logging.basicConfig(format="[%(asctime)s] %(levelname)s %(filename)s: %(message)s", level=level)
    if level <= logging.DEBUG:
        ext_logger.setLevel(logging.WARNING)






import threading
import time

from wifimeter.util.coder import encode
from wifimeter.util.output import output_interface


class UDPSendThread(threading.Thread):
    """
    Sends the message for a measurement request
    :param message_str: message as a json-string; can be send after encoding to the device
    :param socket_object: a object from type UdpClient with an open socket inside
    """
    def __init__(self, message_str, socket_object):
        self._message_str = message_str
        self._udp = socket_object
        super(UDPSendThread, self).__init__()
        self._running = False

    def run(self):
        self._running = True
        while self._running:
            next_send = time.time() + 1
            self.send()
            next_send -= time.time()  # ToDo: not perfect, but it works (probably send a request on every full second)
            time.sleep(next_send)

    def send(self):
        self._udp.send(encode(self._message_str))

    def stop(self):
        self._running = False


class UDPRecvThread(threading.Thread):
    """
    Receives a byte_array from the device on the open socket
    :param socket_object: a object from type UdpClient with an open socket inside
    """
    def __init__(self, socket_object):
        self._udp = socket_object
        super(UDPRecvThread, self).__init__()
        self._running = False

    def run(self):
        self._running = True
        while self._running:
            self.recv()

    def recv(self):
        raw_bytes = self._udp.recv()
        if raw_bytes is not None:
            output_interface(raw_bytes)

    def stop(self):
        self._running = False



from wifimeter.network.tcp import TcpClient
from wifimeter.util.coder import encode, decode
from wifimeter.util.model import SetTimeRequest, SetNetifRequest, NetifRefresher, SetAliasRequest, SetScheduleRequest
import json
import logging


def tcp_setup(wlan_type, wlan_name, alias, password):
    """
    Prepares the data before they can be send to the device
    :param wlan_type: a integer; 0 = without any security, 2 = WEP, 3 = WPA
    :param wlan_name: the name of the required wlan
    :param alias: device name
    :param password: wlan password
    """

    tcp = TcpClient()
    tcp_send(tcp, wlan_type, wlan_name, alias, password)


def tcp_send(tcp, wlan_type, wlan_name, alias, password):
    """
    Sends and receives the TCP-messages for the setup and checks its success
    :param tcp: an object of type TcpClient
    :param wlan_type: a integer; 0 = without security, 2 = WEP, 3 = WPA
    :param wlan_name: the name of the required wlan
    :param alias: device name
    :param password: wlan password
    """
    events = [SetTimeRequest(), SetAliasRequest(alias), NetifRefresher(),
              SetNetifRequest(wlan_name, password, wlan_type), SetScheduleRequest()]
    for req in events:
        tcp.send(encode(req.to_json()))
        decoded_response = json.loads(decode(tcp.recv()))
        logging.debug(decoded_response)
        if not req.validate_response(decoded_response):
            logging.info("invalid err_code")



from setuptools import setup, find_packages

__version__ = "0.1.1"

with open("README.md", "r") as f:
    long_description = f.read()

setup(name='wifimeter',
      version=__version__,
      description='Wifi Electricity Meter',
      url='https://github.com/deddiag/wifimeter.git',
      license='MIT License',
      packages=find_packages(),
      entry_points='''
        [console_scripts]
        wifimeter=wifimeter.cli:cli
      ''',
      install_requires=[
          'pytz',
          'click',
          'timeout_decorator'
      ],
      python_requires='>=2.7',
      long_description_content_type="text/markdown",
      long_description=long_description,
      classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
      ],
)



import socket
import select

UDP_IP = '<broadcast>'
UDP_PORT = 9999
RECV_BYTE_SIZE = 2048


class UdpClient(object):
    """
    opens a socket for every object and close it after using
    an object of the class UdpClient can call the methods send or recv for UDP-communication
    """

    def __init__(self):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

    def __del__(self):
        self._socket.close()

    def send(self, encoded_message):
        self._socket.sendto(encoded_message, (UDP_IP, UDP_PORT))

    def recv(self):
        r, w, e = select.select([self._socket], [], [], 1)
        if self._socket in r:
            recv_bytes, sender_addr = self._socket.recvfrom(RECV_BYTE_SIZE)
            return recv_bytes

    def close(self):
        self._socket.close()



import base64
import unittest

from wifimeter.util.coder import decode, encode


class TestCoder(unittest.TestCase):
    def setUp(self):
        self.message = "{\"system\":{\"get_sysinfo\":null},\"emeter\":{\"get_realtime\":null}}"
        self.byte_array = base64.b64decode("0PKB+Iv/mvfV75S20bTAn+yV5o/hh+jK8J7rh+uWupj9kPWB5Ja0jvXX"
                                           "sNWh/ozpiOSQ+ZTx0+mH8p7yj/I=")

    def test_encode_decode_twice(self):
        result_byte_array = encode(self.message)
        result_after_encode_and_decode = decode(result_byte_array)
        result_byte_array = encode(result_after_encode_and_decode)
        result_after_2_encode_and_decode = decode(result_byte_array)
        self.assertEquals(self.message, result_after_2_encode_and_decode)

    def test_encode(self):
        self.assertEquals(self.byte_array, encode(self.message))

    def test_decode(self):
        self.assertEquals(self.message, decode(self.byte_array))
